from fastapi import FastAPI, Response, Request
from fastapi.responses import PlainTextResponse
from audiobook.servant import FeedServant
from opendal import Operator
from audiobook.config import Config
import tomli
import os
import click


class State:
    config: Config
    op: Operator
    servant: FeedServant


def on_startup():
    with open(os.getenv("AUDIOBOOK_CONFIG", "config.toml"), "rb") as f:
        State.config = Config.model_validate(tomli.load(f))
    State.op = Operator(**State.config.audiobook.opendal)
    State.servant = FeedServant(op=State.op)


app = FastAPI(on_startup=[on_startup])


@app.get("/")
def health():
    return {"Hello": "World"}


@app.get("/books")
def list_books(request: Request):
    return Response(
        "<br>".join(
            [
                f'- <a href="{request.base_url}feed/{book}">{book}</a>'
                for book in State.servant.all_books()
            ]
        ),
        media_type="text/html",
    )


@app.get("/feed/{book}")
def read_feed(book: str, request: Request):
    return Response(
        content=State.servant.get_book_feed(book, str(request.base_url)),
        media_type="application/rss+xml",
    )


@app.get("/file/{book}/{ep_audio}")
def read_ep_file(book: str, ep_audio: str):
    path = f"{book}/{ep_audio}"
    meta = State.op.stat(path)
    content = State.op.read(path)
    return Response(
        content=content,
        media_type=meta.content_type,
    )


@app.get("/text/{book}/{ep_text}")
def read_ep_text(book: str, ep_text: str):
    path = f"{book}/{ep_text}"
    content = State.op.read(path).decode()
    return PlainTextResponse(content)


@click.command()
@click.option("--port", default=8000)
def start(port: int):
    import uvicorn

    uvicorn.run(app, port=port)



import re
import ebooklib
from ebooklib import epub
from bs4 import BeautifulSoup
from pydub import AudioSegment
from typing import Optional, List
from pathlib import Path
from audiobook.config import Config, InferPoolConfig
from pydantic import BaseModel


def chapter_to_str(soup: BeautifulSoup):
    all_text = ""
    for child in soup.children:
        if child.name in {None, "a", "span", "strong", "em"}:
            text = child.get_text(strip=True)
            text = re.sub(r"\[\d+\]", "", text)  # Replace like [1]
            all_text += text
        else:
            all_text += "\n" + chapter_to_str(child)
    return all_text


def extract_title(soup: BeautifulSoup):
    title_tag = soup.find("title")
    if title_tag:
        return title_tag.get_text(separator=",", strip=True)

    for header_tag in ["h1", "h2", "h3", "h4", "h5", "h6"]:
        header = soup.find(header_tag)
        if header:
            return header.get_text(separator=",", strip=True)

    return "Unknown"


def make_filename_safe(input_str):
    cleaned_str = re.sub(r'[\\/*?:"<>|]', " ", input_str)
    cleaned_str = " ".join(cleaned_str.split())[:255]
    return cleaned_str


def load_infer_pool(config: InferPoolConfig):

    from gpt_sovits.infer.inference_pool import GPTSoVITSInferencePool

    path_base = Path(config.path_base)
    with open(path_base / "prompts" / f"{config.prompt_name}.txt") as f:
        prompt_text = f.read().strip()
    return GPTSoVITSInferencePool(
        bert_path=str(path_base / "chinese-roberta-wwm-ext-large"),
        cnhubert_base_path=str(path_base / "chinese-hubert-base"),
        gpt_path=str(path_base / "models" / f"{config.tts_model_name}.ckpt"),
        sovits_path=str(path_base / "models" / f"{config.tts_model_name}.pth"),
        prompt_text=prompt_text,
        prompt_audio_path=str(path_base / "prompts" / f"{config.prompt_name}.wav"),
        device=config.device,
        is_half=config.is_half,
        max_workers=config.max_workers,
    )


class ChapterInfo(BaseModel):
    href: str
    title: str
    text: str


class BookInfo(BaseModel):
    name: str
    chapters: List[ChapterInfo]


def parse_book(book: epub.EpubBook, name: str):
    chapters = list(book.get_items_of_type(ebooklib.ITEM_DOCUMENT))
    cnt_chap = len(chapters)
    max_length = len(str(cnt_chap - 1))
    tmp = [
        (
            idx,
            chapter.file_name,
            BeautifulSoup(chapter.get_body_content(), "html.parser"),
        )
        for idx, chapter in enumerate(chapters)
    ]
    return BookInfo(
        name=name,
        chapters=[
            ChapterInfo(
                href=href,
                text=chapter_to_str(soup),
                title=f"{str(idx).zfill(max_length)}"
                "-"
                f"{make_filename_safe(extract_title(soup))}",
            )
            for idx, href, soup in tmp
        ],
    )


def generate_book(
    book: BookInfo,
    selected_chaps: List[int],
    inferpool_conf: InferPoolConfig,
):
    # Load inference pool

    infer = load_infer_pool(inferpool_conf)

    try:
        for idx in selected_chaps:
            text = book.chapters[idx].text
            sr, data = infer.get_tts_wav(text)
            audio = AudioSegment(
                data=data.tobytes(),
                sample_width=data.dtype.itemsize,
                frame_rate=sr,
                channels=1,
            )
            yield idx, audio
    except Exception as e:
        raise e
    finally:
        infer.pool.shutdown()



import mimetypes
from opendal import Operator, EntryMode
from feedgen.feed import FeedGenerator
from pathlib import Path
from pydantic import BaseModel
from datetime import datetime, timedelta, timezone


class Episode(BaseModel):
    name: str
    extension: str
    path: Path


class FeedServant:
    op: Operator

    def __init__(self, op: Operator):
        self.op = op

    def all_books(self):
        entries = [
            Path(entry.path)
            for entry in self.op.list("/")
            if self.op.stat(entry.path).mode.is_dir()
        ]
        return [entry.stem for entry in entries]

    def get_book_eps(self, book: str):
        extensions = {".mp3"}
        eps = [
            Path(entry.path)
            for entry in self.op.list(f"{book}/")
            if Path(entry.path).suffix in extensions
        ]
        eps.sort(key=lambda x: x.stem)
        return [
            Episode(
                name=entry.stem,
                extension=entry.suffix,
                path=entry,
            )
            for entry in eps
        ]

    def get_book_feed(self, book: str, url_base: str):
        eps = self.get_book_eps(book)
        book_link = f"{url_base}feed/{book}"
        fg = FeedGenerator()
        fg.load_extension("podcast")
        fg.id(book_link)
        fg.title(book)
        fg.description(book)
        fg.link(href=book_link)
        pub_time = datetime.fromtimestamp(0, tz=timezone.utc)
        for ep in eps[::-1]:
            ep_link = f"{url_base}file/{book}/{ep.name}{ep.extension}"
            fe = fg.add_entry(order="append")
            fe.id(ep_link)
            fe.title(ep.name)
            fe.description(
                f'<a href="{url_base}text/{book}/{ep.name}.txt">Origin Text</a>'
            )
            fe.enclosure(ep_link, 0, f"audio/{mimetypes.types_map[ep.extension]}")
            fe.published(pub_time)
            pub_time = pub_time + timedelta(days=1)

        return fg.rss_str(pretty=True)



from ebooklib import epub
import io
import time
from pathlib import Path
import rich.table
import typer
import typer.rich_utils
from audiobook.config import Config
from audiobook.servant import FeedServant
from audiobook.generate import (
    parse_book,
    generate_book,
)
from opendal import Operator
import tomli
from rich.progress import track
from tempfile import NamedTemporaryFile
import rich

app = typer.Typer()


class State:
    config: Config
    op: Operator
    servant: FeedServant


@app.callback()
def before_command(config: str = "config.toml"):
    with open(config, "rb") as f:
        State.config = Config.model_validate(tomli.load(f))
    State.op = Operator(**State.config.audiobook.opendal)
    State.servant = FeedServant(op=State.op)


@app.command()
def create(epubfile: str):
    # Parse epub
    book = parse_book(epub.read_epub(epubfile), Path(epubfile).stem)

    # Select chapters
    console = rich.console.Console()
    table = rich.table.Table("No.", "Name", "Count", "Preview", title="Chapters")
    for idx, info in enumerate(book.chapters):
        preview = info.text.replace(info.title, "").replace("\n", " ")[:15] + "..."
        cnt = len(info.text)
        table.add_row(
            f"[green]{str(idx)}[/green]",
            info.title,
            f"[bold red]{cnt}[/bold red]" if cnt > 20000 else f"{cnt}",
            preview,
        )
    console.print(table)

    user_input = typer.prompt("Select chapters (e.g. 1, 2-3)")
    selected_chaps = parse_int_range(user_input)

    # Generate audio and write to opendal
    for idx, audio in track(
        generate_book(book, selected_chaps, State.config.inferpool),
        total=len(selected_chaps),
    ):
        with io.BytesIO() as in_memory_file:
            audio.export(in_memory_file, format="mp3")
            State.op.write(
                f"{book.name}/{book.chapters[idx].title}.mp3", in_memory_file.read()
            )
        State.op.write(
            f"{book.name}/{book.chapters[idx].title}.txt",
            book.chapters[idx].text.encode(),
        )


@app.command()
def listbook():
    for entry in State.servant.all_books():
        print(entry)


def parse_int_range(input_str: str):
    ranges = input_str.split(",")
    result = []
    for part in ranges:
        if "-" in part:
            start, end = map(int, part.split("-"))
            result.extend(range(start, end + 1))
        else:
            result.append(int(part))
    return result


def main():
    app()


if __name__ == "__main__":
    main()



from typing import List, Dict, Any, Optional
from pydantic import BaseModel, Field


class AudiobookConfig(BaseModel):
    opendal: Dict[str, Any]


class InferPoolConfig(BaseModel):
    path_base: str
    tts_model_name: str = Field(alias="model_name")
    prompt_name: str
    device: Optional[str] = None
    is_half: bool = True
    max_workers: int = 4


class Config(BaseModel):
    audiobook: AudiobookConfig
    inferpool: InferPoolConfig
